# Copyright (c) 2015 Uber Technologies, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

from __future__ import absolute_import

import collections
import functools
import logging

from .enum import enum

log = logging.getLogger('tchannel')


"""Types to represent system events"""
EventType = enum(
    'EventType',
    before_send_request=0x00,
    after_send_request=0x01,
    before_send_response=0x10,
    after_send_response=0x11,
    before_receive_request=0x20,
    after_receive_request=0x21,
    before_receive_response=0x30,
    after_receive_response=0x31,
    after_receive_error=0x41,
    after_send_error=0x42,
    on_exception=0x50,
)


class EventHook(object):
    """provide all event hook interfaces

    Customized Hook should should inherit from EventHook class and implement
    the events' hooks that it wants to listen.

    Example::

        TraceHook(EventHook):
            def before_send_request(self, request):
                ....

    """
    def before_send_request(self, request):
        """Called before any part of a ``CALL_REQ`` message is sent."""
        pass

    def after_send_request(self, request):
        """Not implemented."""
        pass

    def before_send_response(self, response):
        """Not implemented."""
        pass

    def after_send_response(self, response):
        """Called after all of a ``CALL_RESP`` message is sent."""
        pass

    def before_receive_request(self, request):
        """Called after a ``CALL_REQ`` message's arg1 (endpoint) is read."""
        pass

    def after_receive_request(self, request):
        """Not implemented."""
        pass

    def before_receive_response(self, response):
        """Not implemented."""
        pass

    def after_receive_response(self, request, response):
        """Called after a ``CALL_RESP`` message is read."""
        pass

    def after_receive_error(self, request, err):
        """Called after a ``error`` message is read."""
        pass

    def after_send_error(self, err):
        """Called after a ``error`` message is sent."""
        pass

    def on_exception(self, request, err):
        """Called on exceptions within TChannel instance.

        :param request:
            The :py:class:`tchannel.tornado.request.Request` object associated
            with this uncaught exception.

        :param err:
            An instance of the unhandled exception.

            As long as this method is not a :py:func:`coroutine`, it will be
            run in the same exception context as the original error. The
            ``traceback`` module may be useful here.
        """
        pass


class EventEmitter(object):
    def __init__(self):
        self.hooks = collections.defaultdict(lambda: [])

    def register_hook(self, hook, event_type=None):
        """
        If ``event_type`` is provided, then ``hook`` will be called whenever
        that event is fired.

        If no ``event_type`` is specifid, but ``hook`` implements any methods
        with names matching an event hook, then those will be registered with
        their corresponding events. This allows for more stateful, class-based
        event handlers.
        """
        if event_type is not None:
            assert type(event_type) is int, "register hooks with int values"
            return self.hooks[event_type].append(hook)

        for event_type in EventType._fields:
            func = getattr(hook, event_type, None)
            if callable(func):
                event_value = getattr(EventType, event_type)
                self.register_hook(func, event_value)

    def fire(self, event, *args, **kwargs):
        for hook in self.hooks[event]:
            try:
                hook(*args, **kwargs)
            except Exception as e:
                log.error(e.message)


class EventRegistrar(object):

    def __init__(self, event_emitter):
        self.event_emitter = event_emitter

    def register(self, hook, event_type=None):
        return self.event_emitter.register_hook(hook, event_type)

    def __getattr__(self, attr):
        if attr in EventType._fields:
            event_type = getattr(EventType, attr)
            return functools.partial(
                self.register,
                event_type=event_type,
            )
        return super(EventEmitter, self).__getattr__(attr)



# Copyright (c) 2015 Uber Technologies, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

from __future__ import absolute_import

from tchannel.zipkin.trace import Trace


#: The request timed out.
TIMEOUT = 0x01

#: The request was canceled.
CANCELED = 0x02

#: The server was busy.
BUSY = 0x03

# The server declined the request.
DECLINED = 0x04

# The server's handler raised an unexpected exception.
UNEXPECTED_ERROR = 0x05

#: The request was bad.
BAD_REQUEST = 0x06

#: There was a network error when sending the request.
NETWORK_ERROR = 0x07

#: The server handling the request is unhealthy.
UNHEALTHY = 0x08

#: There was a fatal protocol-level error.
FATAL = 0xFF


class TChannelError(Exception):
    """A TChannel-generated exception.

    :ivar code:
        The error code for this error. See the `Specification`_ for a
        description of these codes.
    :vartype code:

    .. _`Specification`:
            http://tchannel.readthedocs.org/en/latest/protocol/#code1_1
    """

    __slots__ = (
        'code',
        'description',
        'id',
        'tracing',
    )

    code = None

    def __init__(
        self,
        description=None,
        id=None,
        tracing=None,
    ):
        super(TChannelError, self).__init__(description)
        self.tracing = tracing or Trace()
        self.id = id
        self.description = description

    @classmethod
    def from_code(cls, code, **kw):
        """Construct a ``TChannelError`` instance from an error code.

        This will return the appropriate class type for the given code.
        """
        return {
            TIMEOUT: TimeoutError,
            CANCELED: CanceledError,
            BUSY: BusyError,
            DECLINED: DeclinedError,
            UNEXPECTED_ERROR: UnexpectedError,
            BAD_REQUEST: BadRequestError,
            NETWORK_ERROR: NetworkError,
            UNHEALTHY: UnhealthyError,
            FATAL: FatalProtocolError,
        }[code](**kw)


class RetryableError(TChannelError):
    """An error where the original request is always safe to retry.

    It is always safe to retry a request with this category of errors. The
    original request was never handled.
    """


class MaybeRetryableError(TChannelError):
    """An error where the original request may be safe to retry.

    The original request may have reached the intended service. Hence, the
    request should only be retried if it is known to be `idempotent`_.

    .. _`idempotent`:
        https://en.wikipedia.org/wiki/Idempotence#Computer_science_meaning
    """


class NotRetryableError(TChannelError):
    """An error where the original request should not be re-sent.

    Something was fundamentally wrong with the request and it should not be
    retried.
    """


class TimeoutError(MaybeRetryableError):
    code = TIMEOUT


class CanceledError(NotRetryableError):
    code = CANCELED


class BusyError(RetryableError):
    code = BUSY


class DeclinedError(RetryableError):
    code = DECLINED


class UnexpectedError(MaybeRetryableError):
    code = UNEXPECTED_ERROR


class BadRequestError(NotRetryableError):
    code = BAD_REQUEST


class NetworkError(MaybeRetryableError):
    code = NETWORK_ERROR


class UnhealthyError(NotRetryableError):
    code = UNHEALTHY


class FatalProtocolError(NotRetryableError):
    code = FATAL


class ReadError(FatalProtocolError):
    """Raised when there is an error while reading input."""
    pass


class InvalidChecksumError(FatalProtocolError):
    """Represent invalid checksum type in the message"""
    pass


class NoAvailablePeerError(RetryableError):
    """Represents a failure to find any peers for a request."""
    pass


class AlreadyListeningError(FatalProtocolError):
    """Raised when attempting to listen multiple times."""
    pass


class OneWayNotSupportedError(BadRequestError):
    """Raised when a one-way Thrift procedure is called."""
    pass


class ValueExpectedError(BadRequestError):
    """Raised when a non-void Thrift response contains no value."""
    pass


class SingletonNotPreparedError(TChannelError):
    """Raised when calling get_instance before calling prepare."""
    pass



# Copyright (c) 2015 Uber Technologies, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

from __future__ import absolute_import

import zlib
from collections import namedtuple

import crcmod.predefined

from .. import rw
from ..enum import enum
from ..errors import InvalidChecksumError
from .types import Types

PROTOCOL_VERSION = 0x02


# 64KB Max frame size
# 16B (size:2 | type:1 | reserved:1 | id:4 | reserved:8)
# 1 2 Bytes can represent 0~2**16-1
MAX_PAYLOAD_SIZE = 0xFFEF   # 64*1024 - 16 - 1


StreamState = enum(
    'StreamState',
    init=0x00,
    streaming=0x01,
    completed=0x02,
    none=0x03,
)


FlagsType = enum(
    'FlagsType',
    none=0x00,
    fragment=0x01,
)


Tracing = namedtuple('Tracing', 'span_id parent_id trace_id traceflags')

tracing_rw = rw.instance(
    Tracing,
    ("span_id", rw.number(8)),      # span_id:8
    ("parent_id", rw.number(8)),    # parent_id:8
    ("trace_id", rw.number(8)),     # trace_id:8
    ("traceflags", rw.number(1)),   # traceflags:1
)


ChecksumType = enum(
    'ChecksumType',
    none=0x00,
    crc32=0x01,
    farm32=0x02,
    crc32c=0x03,
)


checksum_rw = rw.switch(
    rw.number(1),   # csumtype:1
    {
        ChecksumType.none: rw.none(),
        ChecksumType.crc32: rw.number(4),   # csum:4
        ChecksumType.farm32: rw.number(4),  # csum:4
        ChecksumType.crc32c: rw.number(4),  # csum:4
    }
)


CHECKSUM_MSG_TYPES = [Types.CALL_REQ,
                      Types.CALL_REQ_CONTINUE,
                      Types.CALL_RES,
                      Types.CALL_RES_CONTINUE]

# generate crc32c func at import time
crc32c = crcmod.predefined.mkCrcFun('crc-32c')


def compute_checksum(checksum_type, args, csum=0):
    if csum is None:
        csum = 0

    if checksum_type == ChecksumType.none:
        return None
    elif checksum_type == ChecksumType.crc32:
        for arg in args:
            csum = zlib.crc32(arg, csum) & 0xffffffff
    # TODO figure out farm32 cross platform issue
    elif checksum_type == ChecksumType.farm32:
        raise NotImplementedError()
    elif checksum_type == ChecksumType.crc32c:
        for arg in args:
            csum = crc32c(arg, csum)
    else:
        raise InvalidChecksumError()

    return csum


def generate_checksum(message, previous_csum=0):
    """Generate checksum for messages with
        CALL_REQ, CALL_REQ_CONTINUE,
        CALL_RES,CALL_RES_CONTINUE types.

    :param message: outgoing message
    :param previous_csum: accumulated checksum value
    """
    if message.message_type in CHECKSUM_MSG_TYPES:
        csum = compute_checksum(
            message.checksum[0],
            message.args,
            previous_csum,
        )

        message.checksum = (message.checksum[0], csum)


def verify_checksum(message, previous_csum=0):
    """Verify checksum for incoming message.

    :param message: incoming message
    :param previous_csum: accumulated checksum value

    :return return True if message checksum type is None
    or checksum is correct
    """
    if message.message_type in CHECKSUM_MSG_TYPES:
        csum = compute_checksum(
            message.checksum[0],
            message.args,
            previous_csum,
        )

        if csum == message.checksum[1]:
            return True
        else:
            return False
    else:
        return True



# Copyright (c) 2015 Uber Technologies, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

from collections import deque

import tornado
import tornado.concurrent
import tornado.gen
import tornado.ioloop
from tornado.iostream import PipeIOStream
from tornado.iostream import StreamClosedError
from tornado.locks import Condition

from ..errors import UnexpectedError
from ..messages import common
from ..messages.common import StreamState


@tornado.gen.coroutine
def read_full(stream):
    """Read the full contents of the given stream into memory.

    :return:
        A future containing the complete stream contents.
    """
    assert stream, "stream is required"

    chunks = []
    chunk = yield stream.read()

    while chunk:
        chunks.append(chunk)
        chunk = yield stream.read()

    raise tornado.gen.Return(b''.join(chunks))


class Stream(object):

    def read(self):
        """Async read from internal stream buffer

        if it reaches the end of the stream, it will return empty(""), caller
        will depend on the return value to tell if it reaches the end of the
        stream.

        It doesn't support seek functionality, which means it will only read
        the stream data once in single direction.

        NOTE: The data which has been read will be discarded in the stream.

        :return:chunk of bytes read from stream
        """
        raise NotImplementedError()

    def write(self, chunk):
        """Async write to internal stream buffer

        :raises UnexpectedError:
            if stream has been closed, it will raise UnexpectedError
        """
        raise NotImplementedError()

    def set_exception(self, exception):
        """Set exception to interrupt all Stream operations

        :param exception: exception to set
        """
        raise NotImplementedError()

    def close(self):
        raise NotImplementedError()

    def clone(self):
        """Deep clone the current stream"""
        raise NotImplementedError()


class InMemStream(Stream):

    def __init__(self, buf=None, auto_close=True):
        """In-Memory based stream

        :param buf: the buffer for the in memory stream
        """
        self._stream = deque()
        if buf:
            self._stream.append(buf)
        self.state = StreamState.init
        self._condition = Condition()
        self.auto_close = auto_close

        self.exception = None

    def clone(self):
        new_stream = InMemStream()
        new_stream.state = self.state
        new_stream.auto_close = self.auto_close
        new_stream._stream = deque(self._stream)
        return new_stream

    def read(self):

        def read_chunk(future):
            if self.exception:
                future.set_exception(self.exception)
                return future

            chunk = ""

            while len(self._stream) and len(chunk) < common.MAX_PAYLOAD_SIZE:
                chunk += self._stream.popleft()

            future.set_result(chunk)
            return future

        read_future = tornado.concurrent.Future()

        # We're not ready yet
        if self.state != StreamState.completed and not len(self._stream):
            wait_future = self._condition.wait()
            tornado.ioloop.IOLoop.current().add_future(
                wait_future,
                lambda f: f.exception() or read_chunk(read_future)
            )
            return read_future

        return read_chunk(read_future)

    def write(self, chunk):
        if self.exception:
            raise self.exception

        if self.state == StreamState.completed:
            raise UnexpectedError("Stream has been closed.")

        if chunk:
            self._stream.append(chunk)
            self._condition.notify()

        # This needs to return a future to match the async interface.
        r = tornado.concurrent.Future()
        r.set_result(None)
        return r

    def set_exception(self, exception):
        self.exception = exception
        self.close()

    def close(self):
        self.state = StreamState.completed
        self._condition.notify()


class PipeStream(Stream):

    def __init__(self, rpipe, wpipe=None, auto_close=False):
        """Pipe-based stream

        NOTE: reading from or writing to files, use os.open to get the file
        descriptor instead of python's open. Socket file descriptors and
        others are fine.

        when you use os.pipe to generate one write pipe and one read pipe, you
        need to pass both of them into init method.

        :param rpipe: an integer file descriptor which supports read ops
        :param wpipe: an integer file descriptor which supports write ops
        :param auto: flag to indicate to close the stream automatically or not
        """
        assert rpipe is not None
        self._rpipe = rpipe
        self._wpipe = wpipe

        self._rs = (PipeIOStream(self._rpipe) if
                    self._rpipe is not None else None)
        self._ws = (PipeIOStream(self._wpipe) if
                    self._wpipe is not None else None)
        self.auto_close = auto_close
        self.state = StreamState.init

        self.exception = None

    @tornado.gen.coroutine
    def read(self):
        if self.exception:
            raise self.exception

        if self.state == StreamState.completed or self._rpipe is None:
            raise tornado.gen.Return("")
        elif self.state == StreamState.init:
            self.state = StreamState.streaming

        chunk = ""
        try:
            chunk = yield self._rs.read_bytes(
                common.MAX_PAYLOAD_SIZE, partial=True)

        except StreamClosedError:
            # reach the end of the pipe stream
            self.state = StreamState.completed
        finally:
            if self.exception:
                raise self.exception
            raise tornado.gen.Return(chunk)

    @tornado.gen.coroutine
    def write(self, chunk):
        assert self._wpipe is not None
        if self.exception:
            raise self.exception

        try:
            yield self._ws.write(chunk)
            self.state = StreamState.streaming
        except StreamClosedError:
            self.state = StreamState.completed
            raise UnexpectedError("Stream has been closed.")
        finally:
            if self.exception:
                raise self.exception

    def set_exception(self, exception):
        self.exception = exception
        self.close()

    def close(self):
        self.state = StreamState.completed
        if self._ws and self.auto_close:
            self._ws.close()

        if self._rs and self.auto_close:
            self._rs.close()


def maybe_stream(s):
    """Ensure that the given argument is a stream."""
    if isinstance(s, Stream):
        return s

    if s is None:
        stream = InMemStream()
        stream.close()  # we don't intend to write anything
        return stream

    if isinstance(s, unicode):
        s = s.encode('utf-8')
    if isinstance(s, bytearray):
        s = bytes(s)

    if isinstance(s, bytes):
        stream = InMemStream(s)
        stream.close()  # we don't intend to write anything
        return stream

    # s may still conform to the Stream interface. Yay duck typing.
    return s



# Copyright (c) 2015 Uber Technologies, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

from tornado import gen

from . import JSON
from ..serializer.json import JsonSerializer


class JsonArgScheme(object):
    """Semantic params and serialization for json."""

    NAME = JSON

    def __init__(self, tchannel):
        self._tchannel = tchannel

    @gen.coroutine
    def __call__(
        self,
        service,
        endpoint,
        body=None,
        headers=None,
        timeout=None,
        retry_on=None,
        retry_limit=None,
        hostport=None,
        shard_key=None,
        trace=None,
        routing_delegate=None,
    ):
        """Make JSON TChannel Request.

        .. code-block: python

            from tchannel import TChannel

            tchannel = TChannel('my-service')

            resp = tchannel.json(
                service='some-other-service',
                endpoint='get-all-the-crackers',
                body={
                    'some': 'dict',
                },
            )

        :param string service:
            Name of the service to call.
        :param string endpoint:
            Endpoint to call on service.
        :param string body:
            A raw body to provide to the endpoint.
        :param dict headers:
            Dictionary of header key-value pairs.
        :param float timeout:
            How long to wait (in seconds) before raising a ``TimeoutError`` -
            this defaults to ``tchannel.glossary.DEFAULT_TIMEOUT``.
        :param string retry_on:
            What events to retry on - valid values can be found in
            ``tchannel.retry``.
        :param string retry_limit:
            How many times to retry before
        :param string hostport:
            A 'host:port' value to use when making a request directly to a
            TChannel service, bypassing Hyperbahn.
        :param routing_delegate:
            Name of a service to which the request router should forward the
            request instead of the service specified in the call req.

        :rtype: Response
        """

        # serialize
        serializer = JsonSerializer()
        headers = serializer.serialize_header(headers)
        body = serializer.serialize_body(body)

        response = yield self._tchannel.call(
            scheme=self.NAME,
            service=service,
            arg1=endpoint,
            arg2=headers,
            arg3=body,
            timeout=timeout,
            retry_on=retry_on,
            retry_limit=retry_limit,
            hostport=hostport,
            shard_key=shard_key,
            trace=trace,
            routing_delegate=routing_delegate,
        )

        # deserialize
        response.headers = serializer.deserialize_header(response.headers)
        response.body = serializer.deserialize_body(response.body)

        raise gen.Return(response)

    def register(self, endpoint, **kwargs):

        if callable(endpoint):
            handler = endpoint
            endpoint = None
        else:
            handler = None

        return self._tchannel.register(
            scheme=self.NAME,
            endpoint=endpoint,
            handler=handler,
            **kwargs
        )



# Copyright (c) 2015 Uber Technologies, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

from __future__ import absolute_import

from . import common
from .base import BaseMessage
from .common import FlagsType


class CallContinueMessage(BaseMessage):
    """Represent a continuation of a call request (across multiple frames)."""
    max_args_num = 3

    __slots__ = BaseMessage.__slots__ + (
        'flags',
        'checksum',
        'args',
    )

    def __init__(
        self,
        flags=0,
        checksum=None,
        args=None,
        id=0,
    ):
        super(CallContinueMessage, self).__init__(id)
        self.flags = flags

        if checksum is None:
            self.checksum = (common.ChecksumType.none, None)
        else:
            self.checksum = (checksum[0], checksum[1])

        self.args = args or []

    def fragment(self, space_left, fragment_msg):
        """Streaming Message got fragmented based on
        payload size. All the data within space_left
        will be kept. All the rest will be shifted to
        next fragment message.

        :param space_left:
            space left for current frame
        :param fragment_msg:
            the type is either CallRequestMessage or
            CallResponseMessage
        :return: None if there is space left
            or next fragment message
        """
        new_args = []
        key_length = 2  # 2bytes for size
        for i, arg in enumerate(self.args):
            if space_left >= key_length:
                space_left -= key_length

                if arg is not None:
                    arg_length = len(arg)
                    if space_left < arg_length:
                        fragment_msg.args.append(arg[space_left:])
                        new_args.append(arg[:space_left])
                        space_left = 0
                    else:
                        new_args.append(arg)
                        space_left -= arg_length
                        if space_left <= key_length:
                            # boundary for arg
                            fragment_msg.args.append("")
                else:
                    new_args.append("")
            else:
                for l in range(i, len(self.args)):
                    fragment_msg.args.append(self.args[l])
                break

        self.args = new_args
        if space_left >= 0 and len(fragment_msg.args) == 0:
            # don't need to fragment any more
            return None
        else:
            self.flags = FlagsType.fragment
            fragment_msg.id = self.id
            return fragment_msg



# Copyright (c) 2015 Uber Technologies, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

from tchannel.schemes import THRIFT

from .. import io
from .. import rw


class ThriftSerializer(object):

    name = THRIFT

    _headers_rw = rw.headers(
        rw.number(2),
        rw.len_prefixed_string(rw.number(2)),
        rw.len_prefixed_string(rw.number(2)),
    )

    def __init__(self, deserialize_type):
        self.deserialize_type = deserialize_type

    def serialize_header(self, headers):
        headers = headers or {}
        result = self._headers_rw.write(headers, io.BytesIO()).getvalue()
        return result

    def deserialize_header(self, headers):
        headers = headers or {}
        if headers:
            headers = io.BytesIO(headers)
            headers = self._headers_rw.read(headers)
        result = dict(headers)

        return result

    def serialize_body(self, call_args):

        # TODO - use fastbinary directly
        #
        # fastbinary.encode_binary(
        #     call_args, (call_args.__class__, call_args.thrift_spec)
        # )
        # fastbinary.decode_binary(
        #    result, TMemoryBuffer(body),(result_type, result_type.thrift_spec)
        # )
        #
        from thrift.protocol import TBinaryProtocol
        from thrift.transport import TTransport
        trans = TTransport.TMemoryBuffer()
        proto = TBinaryProtocol.TBinaryProtocolAccelerated(trans)
        call_args.write(proto)
        result = trans.getvalue()

        return result

    def deserialize_body(self, body):
        from thrift.protocol import TBinaryProtocol
        from thrift.transport import TTransport
        trans = TTransport.TMemoryBuffer(body)
        proto = TBinaryProtocol.TBinaryProtocolAccelerated(trans)

        result = self.deserialize_type()
        result.read(proto)
        return result


class ThriftRWSerializer(ThriftSerializer):

    def __init__(self, module, deserialize_type):
        """
        :param module:
            thriftrw generated module
        """
        # TODO export dumps/loads on classes in thriftrw in addition to
        # module-level dumps/loads.
        super(ThriftRWSerializer, self).__init__(deserialize_type)
        self.module = module

    def serialize_body(self, obj):
        return self.module.dumps(obj)

    def deserialize_body(self, body):
        return self.module.loads(self.deserialize_type, body)



# Copyright (c) 2015 Uber Technologies, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

from __future__ import absolute_import

import logging
import os
import socket
import sys


import tornado.gen
import tornado.iostream
import tornado.queues as queues

from tornado import stack_context
from tornado.iostream import StreamClosedError

from .. import errors
from .. import frame
from .. import messages
from ..errors import NetworkError
from ..errors import TChannelError
from ..event import EventType
from ..glossary import (
    TCHANNEL_LANGUAGE,
    TCHANNEL_LANGUAGE_VERSION,
    TCHANNEL_VERSION,
    MAX_MESSAGE_ID,
)
from ..io import BytesIO
from ..messages.common import PROTOCOL_VERSION
from ..messages.common import FlagsType
from ..messages.common import StreamState
from ..messages.types import Types
from .message_factory import build_raw_error_message
from .message_factory import MessageFactory
from .util import chain

log = logging.getLogger('tchannel')


#: Sentinel object representing that the connection is outgoing.
OUTGOING = object()

#: Sentinel object representing that the connection is incoming.
INCOMING = object()


class TornadoConnection(object):
    """Manages a bi-directional TChannel conversation between two machines.

    The following primary bi-directional operations are provided:

    ``write(message)``
        Send the message up the wire.
    ``await(message)``
        Receive a message.
    ``send(message)``
        Send a message and receive its response.

    In addition to those, the following operations are provided and should be
    used depending on the direction of the connection.

    ``initiate_handshake``
        Perform a handshake with the remote host.
    ``expect_handshake``
        Expect a handshake request from the remote host.
    """

    CALL_REQ_TYPES = frozenset([Types.CALL_REQ, Types.CALL_REQ_CONTINUE])
    CALL_RES_TYPES = frozenset([Types.CALL_RES, Types.CALL_RES_CONTINUE])

    def __init__(self, connection, tchannel=None, direction=None):
        assert connection, "connection is required"

        self.closed = False
        self.connection = connection
        self.direction = direction or INCOMING

        sockname = connection.socket.getsockname()
        if len(sockname) == 2:
            (self.remote_host,
             self.remote_host_port) = sockname
        elif len(sockname) == 1:
            self.remote_host = sockname[0]
            self.remote_host_port = 0
        else:
            self.remote_host = "0.0.0.0"
            self.remote_host_port = 0

        self.remote_host_port = int(self.remote_host_port)
        self.remote_process_name = None
        self.requested_version = PROTOCOL_VERSION

        # Tracks message IDs for this connection.
        self._id_sequence = 0

        # We need to use two separate message factories to avoid message ID
        # collision while assembling fragmented messages.
        self.request_message_factory = MessageFactory(self.remote_host,
                                                      self.remote_host_port)
        self.response_message_factory = MessageFactory(self.remote_host,
                                                       self.remote_host_port)

        # Queue of unprocessed incoming calls.
        self._messages = queues.Queue()

        # Map from message ID to futures for responses of outgoing calls.
        self._outstanding = {}

        # Whether _loop is running. The loop doesn't run until after the
        # handshake has been performed.
        self._loop_running = False

        self.tchannel = tchannel
        self._close_cb = None

        connection.set_close_callback(self._on_close)

    def set_close_callback(self, cb):
        """Specify a function to be called when this connection is closed.

        :param cb:
            A callable that takes no arguments. This callable will be called
            when this connection is closed.
        """
        assert self._close_cb is None, (
            'A close_callback has already been set for this connection.'
        )
        self._close_cb = stack_context.wrap(cb)

    def next_message_id(self):
        self._id_sequence = (self._id_sequence + 1) % MAX_MESSAGE_ID
        return self._id_sequence

    def _on_close(self):
        self.closed = True

        for message_id, future in self._outstanding.iteritems():
            future.set_exception(
                NetworkError(
                    "canceling outstanding request %d" % message_id
                )
            )
        self._outstanding = {}

        try:
            while True:
                message = self._messages.get_nowait()
                log.warn("Unconsumed message %s", message)
        except queues.QueueEmpty:
            pass

        if self._close_cb:
            self._close_cb()

    def await(self):
        """Get the next call to this TChannel."""
        if self._loop_running:
            return self._messages.get()
        else:
            return self._recv()

    def _recv(self):
        """Receive the next message off the wire.

        :returns:
            A Future that produces a Context object containing the next
            message off the wire.
        """

        # This is the message_future we'll return for any inbound messages.
        message_future = tornado.gen.Future()

        def on_body(read_body_future, size):
            if read_body_future.exception():
                return on_error(read_body_future)

            body = read_body_future.result()
            f = frame.frame_rw.read(BytesIO(body), size=size)
            message_rw = messages.RW[f.header.message_type]
            message = message_rw.read(BytesIO(f.payload))
            message.id = f.header.message_id
            message_future.set_result(message)

        def on_read_size(read_size_future):
            if read_size_future.exception():
                return on_error(read_size_future)

            size_bytes = read_size_future.result()
            size = frame.frame_rw.size_rw.read(BytesIO(size_bytes))
            read_body_future = self.connection.read_bytes(size - size_width)

            tornado.ioloop.IOLoop.current().add_future(
                read_body_future,
                lambda future: on_body(future, size)
            )

            return read_body_future

        def on_error(future):
            log.info("Failed to read data: %s", future.exception())

        size_width = frame.frame_rw.size_rw.width()
        read_bytes_future = self.connection.read_bytes(size_width)

        tornado.ioloop.IOLoop.current().add_future(
            read_bytes_future,
            on_read_size,
        )

        return message_future

    @tornado.gen.coroutine
    def _loop(self):
        # Receive messages off the wire. All messages are either responses to
        # outstanding requests or calls.
        #
        # Must be started only after the handshake has been performed.
        self._loop_running = True
        while not self.closed:
            message = yield self._recv()

            # TODO: There should probably be a try-catch on the yield.
            if message.message_type in self.CALL_REQ_TYPES:
                self._messages.put(message)
                continue

            elif message.id in self._outstanding:
                # set exception if receive error message
                if message.message_type == Types.ERROR:
                    future = self._outstanding.pop(message.id)
                    if future.running():
                        error = TChannelError.from_code(
                            message.code,
                            description=message.description,
                            id=message.id,
                            tracing=message.tracing,
                        )
                        future.set_exception(error)
                    else:
                        protocol_exception = (
                            self.response_message_factory.build(message)
                        )
                        if protocol_exception:
                            self.event_emitter.fire(
                                EventType.after_receive_error,
                                protocol_exception,
                            )
                    continue

                response = self.response_message_factory.build(message)

                # keep continue message in the list
                # pop all other type messages including error message
                if (message.message_type in self.CALL_RES_TYPES and
                        message.flags == FlagsType.fragment):
                    # still streaming, keep it for record
                    future = self._outstanding.get(message.id)
                else:
                    future = self._outstanding.pop(message.id)

                if response and future.running():
                    future.set_result(response)
                continue

            log.warn('Unconsumed message %s', message)

    # Basically, the only difference between send and write is that send
    # sets up a Future to get the response. That's ideal for peers making
    # calls. Peers responding to calls must use write.
    def send(self, message):
        """Send the given message up the wire.
        Use this for messages which have a response message.

        :param message:
            Message to send
        :returns:
            A Future containing the response for the message
        """
        assert self._loop_running, "Perform a handshake first."
        assert message.message_type in self.CALL_REQ_TYPES, (
            "Message '%s' can't use send" % repr(message)
        )

        message.id = message.id or self.next_message_id()
        assert message.id not in self._outstanding, (
            "Message ID '%d' already being used" % message.id
        )

        future = tornado.gen.Future()
        self._outstanding[message.id] = future
        self.write(message)
        return future

    def write(self, message):
        """Writes the given message up the wire.

        Does not expect a response back for the message.

        :param message:
            Message to write.
        """
        message.id = message.id or self.next_message_id()

        if message.message_type in self.CALL_REQ_TYPES:
            message_factory = self.request_message_factory
        else:
            message_factory = self.response_message_factory

        fragments = message_factory.fragment(message)

        return chain(fragments, self._write)

    def _write(self, message):
        """Writes the given message up the wire.

        The message must be small enough to fit in a single frame.
        """
        message.id = message.id or self.next_message_id()

        payload = messages.RW[message.message_type].write(
            message, BytesIO()
        ).getvalue()

        f = frame.Frame(
            header=frame.FrameHeader(
                message_type=message.message_type,
                message_id=message.id,
            ),
            payload=payload
        )
        body = frame.frame_rw.write(f, BytesIO()).getvalue()

        return self.connection.write(body)

    def close(self):
        if not self.closed:
            self.connection.close()

    @tornado.gen.coroutine
    def initiate_handshake(self, headers):
        """Initiate a handshake with the remote host.

        :param headers:
            A dictionary of headers to send.
        :returns:
            A future that resolves (with a value of None) when the handshake
            is complete.
        """
        self._write(messages.InitRequestMessage(
            version=PROTOCOL_VERSION,
            headers=headers
        ))
        init_res = yield self._recv()
        if init_res.message_type != Types.INIT_RES:
            raise errors.InvalidMessageError(
                "Expected handshake response, got %s" % repr(init_res)
            )
        self._extract_handshake_headers(init_res)

        # The receive loop is started only after the handshake has been
        # completed.
        self._loop()

        raise tornado.gen.Return(init_res)

    @tornado.gen.coroutine
    def expect_handshake(self, headers):
        """Expect a handshake from the remote host.

        :param headers:
            Headers to respond with
        :returns:
            A future that resolves (with a value of None) when the handshake
            is complete.
        """
        init_req = yield self._recv()
        if init_req.message_type != Types.INIT_REQ:
            raise errors.InvalidMessageError(
                "You need to shake my hand first. Got %s" % repr(init_req)
            )
        self._extract_handshake_headers(init_req)

        self._write(
            messages.InitResponseMessage(
                PROTOCOL_VERSION, headers, init_req.id),
        )

        # The receive loop is started only after the handshake has been
        # completed.
        self._loop()

        raise tornado.gen.Return(init_req)

    def _extract_handshake_headers(self, message):
        if not message.host_port:
            raise errors.InvalidMessageError(
                'Missing required header: host_port'
            )

        if not message.process_name:
            raise errors.InvalidMessageError(
                'Missing required header: process_name'
            )

        (self.remote_host,
         self.remote_host_port) = message.host_port.rsplit(':', 1)
        self.remote_host_port = int(self.remote_host_port)
        self.remote_process_name = message.process_name
        self.requested_version = message.version

    @classmethod
    @tornado.gen.coroutine
    def outgoing(cls, hostport, process_name=None, serve_hostport=None,
                 handler=None, tchannel=None):
        """Initiate a new connection to the given host.

        :param hostport:
            String in the form ``$host:$port`` specifying the target host
        :param process_name:
            Process name of the entity making the connection.
        :param serve_hostport:
            String in the form ``$host:$port`` specifying an address at which
            the caller can be reached. If omitted, ``0.0.0.0:0`` is used.
        :param handler:
            If given, any calls received from this connection will be sent to
            this RequestHandler.
        """
        host, port = hostport.rsplit(":", 1)
        process_name = process_name or "%s[%s]" % (sys.argv[0], os.getpid())
        serve_hostport = serve_hostport or "0.0.0.0:0"

        # TODO: change this to tornado.tcpclient.TCPClient to do async DNS
        # lookups.
        stream = tornado.iostream.IOStream(
            socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        )

        log.debug("Connecting to %s", hostport)
        try:
            yield stream.connect((host, int(port)))
        except (StreamClosedError, socket.error) as e:
            log.warn("Couldn't connect to %s", hostport)
            raise NetworkError(
                "Couldn't connect to %s" % hostport, e
            )

        connection = cls(stream, tchannel, direction=OUTGOING)
        log.debug("Performing handshake with %s", hostport)
        yield connection.initiate_handshake(headers={
            'host_port': serve_hostport,
            'process_name': process_name,
            'tchannel_language': TCHANNEL_LANGUAGE,
            'tchannel_language_version': TCHANNEL_LANGUAGE_VERSION,
            'tchannel_version': TCHANNEL_VERSION,
        })

        if handler:
            connection.serve(handler)

        raise tornado.gen.Return(connection)

    @tornado.gen.coroutine
    def serve(self, handler):
        """Serve calls over this connection using the given RequestHandler.

        :param handler:
            RequestHandler to process the requests through
        :return:
            A Future that resolves (to None) once the loop is done running --
            which happens once this connection is closed.
        """
        assert handler, "handler is required"
        assert self._loop_running, "Finish the handshake first"

        while not self.closed:
            message = yield self.await()

            try:
                handler(message, self)
            except Exception:
                # TODO Send error frame back
                logging.exception("Failed to process %s", repr(message))

    def send_error(self, error):
        """Convenience method for writing Error frames up the wire.

        :param error:
            TChannel Error. :py:class`tchannel.errors.TChannelError`.
        :returns:
            A future that resolves when the write finishes.
        """

        error_message = build_raw_error_message(error)
        write_future = self._write(error_message)
        write_future.add_done_callback(
            lambda f: self.tchannel.event_emitter.fire(
                EventType.after_send_error,
                error,
            )
        )
        return write_future

    def ping(self):
        return self._write(messages.PingRequestMessage())

    def pong(self):
        return self._write(messages.PingResponseMessage())


class StreamConnection(TornadoConnection):
    """Streaming request/response into protocol messages and sent by tornado
    connection

    Here are public apis provided by StreamConnection:
    "post_response(response)"
        stream response object into wire

    "stream_request(request)"
        stream request object into wire without waiting for a response

    "send_request(request)"
        stream request object into wire and wait for a response

    """

    @tornado.gen.coroutine
    def _stream(self, context, message_factory):
        """write request/response into frames

        Transform request/response into protocol level message objects based on
        types and argstreams.

        Assumption: the chunk data read from stream can fit into memory.

        If arg stream is at init or streaming state, build the message based on
        current chunk. If arg stream is at completed state, put current chunk
        into args[] array, and continue to read next arg stream in order to
        compose a larger message instead of sending multi small messages.

        Note: the message built at this stage is not guaranteed the size is
        less then 64KB.

        Possible messages created sequence:

        Take request as an example::
        CallRequestMessage(flags=fragment)
            --> CallRequestContinueMessage(flags=fragment)
            ....
            --> CallRequestContinueMessage(flags=fragment)
                --> CallRequestMessage(flags=none)

        :param context: Request or Response object
        """
        args = []
        try:
            for argstream in context.argstreams:
                chunk = yield argstream.read()
                args.append(chunk)
                chunk = yield argstream.read()
                while chunk:
                    message = (message_factory.
                               build_raw_message(context, args))
                    yield self.write(message)
                    args = [chunk]
                    chunk = yield argstream.read()

            # last piece of request/response.
            message = (message_factory.
                       build_raw_message(context, args, is_completed=True))
            yield self.write(message)
            context.state = StreamState.completed
        # Stop streamming immediately if exception occurs on the handler side
        except TChannelError as e:
            # raise by tchannel intentionally
            log.info("Stop Outgoing Streams because of error: %s", e.message)

    @tornado.gen.coroutine
    def post_response(self, response):
        try:
            # TODO: before_send_response
            yield self._stream(response, self.response_message_factory)

            # event: send_response
            self.tchannel.event_emitter.fire(
                EventType.after_send_response,
                response,
            )
        finally:
            response.close_argstreams(force=True)

    def stream_request(self, request):
        """send the given request and response is not required"""
        request.close_argstreams()

        stream_future = self._stream(request, self.request_message_factory)

        tornado.ioloop.IOLoop.current().add_future(
            stream_future,
            lambda f: request.close_argstreams(force=True)
        )

        return stream_future

    def send_request(self, request):
        """Send the given request and response is required.

        Use this for messages which have a response message.

        :param request:
            request to send
        :returns:
            A Future containing the response for the request
        """
        assert self._loop_running, "Perform a handshake first."

        assert request.id not in self._outstanding, (
            "Message ID '%d' already being used" % request.id
        )

        future = tornado.gen.Future()
        self._outstanding[request.id] = future
        self.stream_request(request)

        # the actual future that caller will yield
        response_future = tornado.gen.Future()
        # TODO: fire before_receive_response

        tornado.ioloop.IOLoop.current().add_future(
            future,
            lambda f: self.adapt_result(f, request, response_future),
        )
        return response_future

    def adapt_result(self, f, request, response_future):
        if not response_future.running():
            return

        if f.exception():
            protocol_exception = f.exception()
            protocol_exception.tracing = request.tracing
            response_future.set_exception(protocol_exception)

        else:
            response = f.result()
            response.tracing = request.tracing
            response_future.set_result(response)

    def remove_outstanding_request(self, request):
        """Remove request from pending request list"""
        self._outstanding.pop(request.id, None)



# Copyright (c) 2015 Uber Technologies, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

from __future__ import absolute_import

from .. import rw
from .init_request import InitRequestMessage
from .types import Types


class InitResponseMessage(InitRequestMessage):
    """Respond to an initialization request message."""
    message_type = Types.INIT_RES


init_res_rw = rw.instance(
    InitResponseMessage,
    ('version', rw.number(2)),  # version:2
    ('headers', rw.headers(     # nh:2 (key~2 value~2){nh}
        rw.number(2),
        rw.len_prefixed_string(rw.number(2)),
        rw.len_prefixed_string(rw.number(2)),
    )),
)



# Copyright (c) 2015 Uber Technologies, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

from __future__ import absolute_import

import contextlib2
import inspect
import wrapt
import sys

from .cassette import Cassette
from .patch import Patcher, force_reset
from .server import VCRProxyService


class _CassetteContext(object):
    """Lets use_cassette be used as a context manager and a decorator."""

    def __init__(self, path, record_mode, inject, matchers):
        self.path = path
        self.record_mode = record_mode
        self.inject = inject
        self.matchers = matchers

        self._exit_stack = contextlib2.ExitStack()

    def __enter__(self):
        cassette = self._exit_stack.enter_context(
            Cassette(
                path=self.path,
                record_mode=self.record_mode,
                matchers=self.matchers,
            )
        )

        server = self._exit_stack.enter_context(
            VCRProxyService(cassette=cassette, unpatch=force_reset)
        )

        # TODO Maybe instead of using this instance of the TChannel client, we
        # should use the one being patched to make the requests?

        self._exit_stack.enter_context(Patcher(server.hostport))

        return cassette

    def __exit__(self, *args):
        self._exit_stack.__exit__(*args)

    def _handle_coroutine(self, function, args, kwargs):
        with self as cassette:
            if self.inject:
                coroutine = function(cassette, *args, **kwargs)
            else:
                coroutine = function(*args, **kwargs)

            # Spin the generator, yielding its results and sending back
            # respnoses and exceptions until it is exhausted. StopIteration
            # will be raised and caught by the caller (@gen.coroutine).
            future = next(coroutine)
            while True:
                try:
                    result = yield future
                except Exception:
                    future = coroutine.throw(*sys.exc_info())
                else:
                    future = coroutine.send(result)

    @wrapt.decorator
    def __call__(self, function, instance, args, kwargs):
        if inspect.isgeneratorfunction(function):
            return self._handle_coroutine(function, args, kwargs)

        with self as cassette:
            if self.inject:
                return function(cassette, *args, **kwargs)
            else:
                return function(*args, **kwargs)


def use_cassette(path, record_mode=None, inject=False, matchers=None):
    """Use or create a cassette to record/replay TChannel requests.

    This may be used as a context manager or a decorator.

    .. code-block:: python

        from tchannel.testing import vcr

        @pytest.mark.gen_test
        @vcr.use_cassette('tests/data/bar.yaml')
        def test_bar():
            channel = TChannel('test-client')
            service_client = MyServiceClient(channel)

            yield service_client.myMethod()


        def test_bar():
            with vcr.use_cassette('tests/data/bar.yaml', record_mode='none'):
                # ...

    Note that when used as a decorator on a coroutine, the ``use_cassette``
    decorator must be applied BEFORE ``gen.coroutine`` or
    ``pytest.mark.gen_test``.

    :param path:
        Path to the cassette. If the cassette did not already exist, it will
        be created. If it existed, its contents will be replayed (depending on
        the record mode).
    :param record_mode:
        The record mode dictates whether a cassette is allowed to record or
        replay interactions. This may be a string specifying the record mode
        name or an element from the
        :py:class:`tchannel.testing.vcr.RecordMode` object. This parameter
        defaults to :py:attr:`tchannel.testing.vcr.RecordMode.ONCE`. See
        :py:class:`tchannel.testing.vcr.RecordMode` for details on supported
        record modes and how to use them.
    :param inject:
        If True, when ``use_cassette`` is used as a decorator, the cassette
        object will be injected into the function call as the first argument.
        Defaults to False.
    :param matchers:
        Used to configure the request attributes which VCR matches on. A
        recorded response will be replayed if all specified attributes of the
        corresponding request match the request that is being made. Valid
        attributes are: ``serviceName``, ``hostPort``, ``endpoint``,
        ``headers``, ``body``, ``argScheme``, and ``transportHeaders``.

        For example,

        .. code-block:: python

            MY_MATCHERS = list(vcr.DEFAULT_MATCHERS)
            MY_MATCHERS.remove('headers')

            @vcr.use_cassette('tests/data/foo.yaml', matchers=MY_MATCHERS):
            def test_foo():
                # ...

        By default, the following attributes are matched: ``serviceName``,
        ``endpoint``, ``headers``, ``body``, and ``argScheme``.
        :py:data:`tchannel.testing.vcr.DEFAULT_MATCHERS` is a tuple of all
        these matchers.
    """

    return _CassetteContext(
        path=path,
        record_mode=record_mode,
        inject=inject,
        matchers=matchers,
    )

    # TODO create some sort of configurable VCR object which implements
    # use_cassette. Top-level use_cassette can just use a default instance.


__all__ = ['use_cassette']
